unit FlipperkastNul;
  { (c) 2002, Tom Verhoeff, TUE }
  { Spelregels: Start de flipperkast, voer een aantal Flips uit en
    wijs de 0-bal aan.  (Oefen-versie) }

interface
  { met specificaties/contracten }

const
  AantalGaten = 10; { aantal gaten in de flipperkast }

type
  GatNummer = 0 .. AantalGaten - 1; { de nummers van de gaten }

procedure Start ( const sNummer     : Integer;
                  out   aantalBallen: Integer );
  { Start de flipperkast en ontvang een aantal ballen. }
  { pre:  Nog geen enkele procedure uit deze unit is aangeroepen.
          sNummer = uw identiteitsnummer (bij wijze van betaling).
    post: aantalBallen = aantal ballen in de flipperkast.
          1 <= aantalBallen <= AantalGaten.
          De gaten 0 t/m aantalBallen - 1 bevatten elk een bal,
          waarop een uniek getal van 0 t/m aantalBallen - 1. }

procedure Flip ( const i, j: GatNummer );
  { Voer een flip uit. }
  { pre:  De flipperkast staat aan (Start is al aangeroepen).
          i, j < aantalBallen (met aantalBallen uit Start).
          Zij  a  het getal op de bal in gat i.
          Zij  b  het getal op de bal in gat j.
    post: Gat i bevat de bal met daarop de kleinste van a en b.
          Gat j bevat de bal met daarop de grootste van a en b.
          De ballen in de andere gaten zijn onaangeroerd. }

procedure Stop ( const gatMetBal0: GatNummer );
  { Wijs de 0-bal aan en stop de flipperkast. }
  { pre:  De flipperkast staat aan (Start is al aangeroepen).
          gatMetBal0 < aantalBallen (uit Start).
          Gat gatMetBal0 bevat de bal met daarop getal 0.
    post: Als antwoord correct is, dan is beloning uitgekeerd.
          Programma is gestopt na lezen van <return>. } 

implementation
  { Het is absoluut niet nodig dit te lezen of te begrijpen. }

const
  ERROR = ' ERROR: ';
  MinStudentNummer = 100000;
  MaxStudentNummer = 670000;

var
  initialized: Boolean;
  wieSpeelt: Integer;
  nBallen: Integer; { aantal ballen }
  balInGat: array [ GatNummer ] of Integer;
    { balInGat [ i ] = nummer op bal in gat i, voor 0 <= i < nBallen }
  nFlips: Integer; { aantal Flip aanroepen }

procedure DelayedHalt ( double: Boolean; code: Integer );
  { stop after reading return; exit with code }
  begin
    if code <> 0 then begin
      writeln ( 'Voortijdige beeindiging vanwege contractbreuk' )
    end { if }
  ; write ( 'Tik <return> om stop te voltooien: ' )
  ; readln
  ; if double then begin
      readln  { just in case of I/O error }
    end { if }
  ; Halt ( code )
  end; { DelayedHalt }

function  ValidStudentNumber ( sNummer: Integer ): Boolean;
  { pre:  True
    ret:  of sNummer een valide TUE studentnummer is }
  var
    TBL: array [ 0 .. 20 ] of Integer; { TBL[i] = 2^1 mod 11 }
    i: Integer; { om cijfers te doorlopen }
    s: Integer; { (+ i: 0 <= i < 6: TBL[6+i+a[i]]), where
                  sNummer = a[0]..a[5] }
  begin
    s := 1 { s = 2^i mod 11 }

//; write ( 'TBL =' )
  ; for i := 0 to 20 do begin
      TBL [ i ] := s
    ; s := ( 2 * s ) mod 11
//; write ( ' ', TBL[i] )
    end { for i }
//; writeln

  ; Result := False  { anticipated }
  ; if ( sNummer < MinStudentNummer ) or ( sNummer > MaxStudentNummer ) then
      Exit
  ; s := 0

  ; for i := 5 downto 0 do begin
      s := s + TBL [ 6 + i + sNummer mod 10 ]
    ; sNummer := sNummer div 10
    end { for i }

  ; Result := ( s mod 10 = 0 )
  end; { ValidStudentNumber }

procedure WriteBalls ( const msg: String );
  var
    gat: GatNummer; { doorloopt gaten met ballen }
  begin
  ; write ( msg, ': balrij =' )

  ; for gat := 0 to nBallen - 1 do begin
      write ( ' ', balInGat [ gat ] )
    end { for gat }

  ; writeln
  end; { WriteBalls }

procedure Start ( const sNummer     : Integer;
                  out   aantalBallen: Integer );
  const
    RoutineName = 'Start';
  var
    gat: GatNummer; { doorloopt gaten met ballen }
    bal: Integer;   { om balnummer in te lezen }
    ior: Word; { om IOResult in op te vangen }
    freeballs: Set of 0 .. AantalGaten - 1; { verzameling beschikbare ballen }
  begin
    if initialized then begin
      writeln ( RoutineName, ERROR,
                'Start mag maar een keer worden aangeroepen.' )
    ; DelayedHalt ( False, 1 )
    end { if }
    { not initialized }

  ; if not ValidStudentNumber ( sNummer) then begin
      writeln ( RoutineName, ERROR,
                sNummer, ' is geen acceptabel TU/e studentnummer' )
    ; DelayedHalt ( False, 1 )
    end { if }
    { sNummer is valide }

  ; initialized := True
  ; wieSpeelt := sNummer
  ; writeln ( RoutineName, ': Welkom TU/e student ', sNummer )
  ; write ( RoutineName,
            ': Hoeveel ballen [ 1 .. ', AantalGaten : 1, ' ]? ' )
    {$I-}
  ; readln ( nBallen )
  ; ior := IOResult
    {$I+}
  ; if ior <> 0 then begin
{ should first clear input buffer ... }
      writeln ( RoutineName, ERROR,
                'I/O error ', ior : 1, ' bij lezen van aantal ballen' )
    ; DelayedHalt ( True, 1 )
    end { if }
  ; if ( nBallen <  1 ) or ( nBallen > AantalGaten ) then begin
      writeln ( RoutineName, ERROR,
                'ongeldig aantal ballen, niet in [ 1 .. ',
                AantalGaten : 1, ' ]' )
    ; DelayedHalt ( True, 1 )
    end { if }
  ; aantalBallen := nBallen
  ; writeln ( RoutineName, ': aantalBallen = ', aantalBallen )
  ; write ( RoutineName,
            ': Getallen op de ballen in gaten 0 t/m ', nBallen - 1, '? ' )
  ; freeballs := [ 0 .. nBallen - 1 ]

  ; for gat := 0 to nBallen - 1 do begin
      {$I-,R-}
      read ( bal )
    ; ior := IOResult
      {$I+,R+}
    ; if ( ior <> 0 ) then begin
        writeln ( RoutineName, ERROR,
                  'I/O error ', ior : 1,
                  ' bij lezen van bal in gat ', gat : 1 )
      ; DelayedHalt ( True, 1 )
      end { if }
    ; if ( bal < 0 ) or ( bal >= nBallen ) then begin
        writeln ( RoutineName, ERROR,
                  'ongeldig balnummer ', bal : 1, ' in gat ', gat : 1,
                  ', niet in [ 0 .. ', nBallen - 1 : 1, ' ]' )
      ; DelayedHalt ( True, 1 )
      end { if }
    ; if not ( bal in freeballs ) then begin
        writeln ( RoutineName, ERROR,
                  'bal ', bal : 1,
                  ' in gat ', gat : 1, ' komt al eerder voor')
      ; DelayedHalt ( True, 1 )
      end { if }
    ; freeballs := freeballs - [ bal ]
    ; balInGat [ gat ] := bal
    end { for gat }

  ; readln

  ; nFlips := 0
  ; WriteBalls ( RoutineName )
  ; writeln
  end; { Start }

procedure Flip ( const i, j: GatNummer );
  const
    RoutineName = 'Flip';
  var
    hulpBal: Integer; { wisselhulp }
  begin
    if not initialized then begin
      writeln ( RoutineName, ERROR,
                'Eerst de flipperkast aanzetten met Start' )
    ; DelayedHalt ( False, 1 )
    end { if }
    { initialized }

  ; writeln ( RoutineName, ': i, j = ', i, ', ', j )
  ; if ( i < 0 ) or ( i >= nBallen ) then begin 
      writeln ( RoutineName, ERROR,
                'i niet in [ 0 .. ', nBallen - 1 : 1, ' ]' )
    ; DelayedHalt ( False, 1 )
    end { if }
    { i < nBallen }

  ; if ( j < 0 ) or ( j >= nBallen ) then begin 
      writeln ( RoutineName, ERROR,
                'j niet in [ 0 .. ', nBallen - 1 : 1, ' ]' )
    ; DelayedHalt ( False, 1 )
    end { if }
    { j < nBallen }

  ; if balInGat [ i ] >  balInGat [ j ] then begin
      { verwissel balInGat[i] en balInGat[j] }
      hulpBal        := balInGat [ i ]
    ; balInGat [ i ] := balInGat [ j ]
    ; balInGat [ j ] := hulpBal
    end { if }
  ; Inc ( nFlips )
  ; WriteBalls ( RoutineName )
  ; writeln
  end; { Flip }

procedure Stop ( const gatMetBal0: GatNummer );
  const
    RoutineName = 'Stop';
  begin
    if not initialized then begin
      writeln ( RoutineName, ERROR,
                'Eerst de flipperkast aanzetten met Start' )
    ; DelayedHalt ( False, 1 )
    end { if }
    { initialized }

  ; writeln ( RoutineName, ': gatMetBal0 = ', gatMetBal0 )
  ; if gatMetBal0 >= nBallen then begin 
      writeln ( RoutineName, ERROR,
                'gatMetBal0 > aantalBallen = ', nBallen : 1 )
    ; DelayedHalt ( False, 1 )
    end { if }

  ; write ( RoutineName, ': Dat is ... <tromgeroffel> ... ' )
  ; if balInGat [ gatMetBal0 ] = 0 then
      writeln ( 'goed! <Muziekje en applaus>' )
    else
      writeln ( 'fout! <Herrie en boegeroep>' )

  ; if nFlips < nBallen - 1 then
      writeln ( RoutineName, ': Je hebt gegokt' )
    else if nFlips > nBallen - 1 then
      writeln ( RoutineName, ': Kan met minder flips' )
  ; DelayedHalt ( False, 0 )
  end; { FiniBalans }

begin
  initialized := False
end.
